/*
    SPDX-FileCopyrightText: 2024 Joshua Goins <josh@redstrate.com>

    SPDX-License-Identifier: MIT
*/
#include "test16.h"

int main(int, char **)
{
    QStandardPaths::setTestModeEnabled(true);
    return 0;
}
