/*
    This file is part of the KDE libraries
    SPDX-FileCopyrightText: 1997 Stefan Taferner <taferner@alpin.or.at>
    SPDX-FileCopyrightText: 2000 Nicolas Hadacek <haadcek@kde.org>
    SPDX-FileCopyrightText: 2001, 2002 Ellis Whitehead <ellis@kde.org>

    SPDX-License-Identifier: LGPL-2.0-or-later
*/

#ifndef KSTANDARDSHORTCUT_P_H
#define KSTANDARDSHORTCUT_P_H

#include "kstandardshortcut.h"

namespace KStandardShortcut
{
void initialize(StandardShortcut id);
}

#endif
